#include <cstdio>
#include <iostream>
#include <fstream>
#include <string>
#include <cctype>
#include <vector>
#include <algorithm>
#include <list>

using namespace std;

static string dest;
static vector<string> srcs;
static unsigned valuecount;
static string hfilename;

static const unsigned MAXFIELDS = 2;

static const string arrname()
{
    string s;
    for(unsigned a=0; a<srcs.size(); ++a)
    {
        s += srcs[a];
        s += '_';
    }
    s += dest;
    return s;
};

struct elem
{
    string desc;
    vector<unsigned> values;
};

typedef vector<elem> lista;

class verta
{
    const lista &values;
    unsigned key;
public:
    verta(const lista &v, unsigned k) : values(v), key(k) { }
    
    bool operator() (unsigned a, unsigned b) const
    {
        return values[a].values[key] < values[b].values[key];
    }        
};

struct hashmath
{
    int integer;        /* Number this is multiplied by      */
    unsigned exponent;  /* 0=constant, 1=x, 2=xx, 3=xxx, ... */
public:
    hashmath() : integer(0), exponent(0) { }
    explicit hashmath(int add, unsigned exp=0) : integer(add), exponent(exp) { }
};

struct rep
{
    unsigned keystart;
    unsigned len;
    
    vector<hashmath> components;
};

struct hashop
{
    bool istable;
    unsigned last;     /* if(n<limit) ...          */
    
    /* With table: */
    int first;    /* table index */
    unsigned dec; /* table value offset */
    
    /* With math: */
    vector<hashmath> components;
};

#if 0
/* Recursive determinant calculation */
static int calcdet(const vector<int> &mat, unsigned depf)
{
    int sum;
    if(depf <= 1)
        sum = mat[0];
    else
    {
        sum = 0;
        for(unsigned a=0; a<depf; ++a)
        {
            unsigned subdepf = depf-1;
            vector<int> tmp(subdepf * subdepf);
            for(unsigned b=0; b<subdepf; ++b)
                for(unsigned c=0; c<subdepf; ++c)
                    tmp[b*subdepf + c] = mat[(1+b) * depf + (c + (c>=a))%depf];
            int subdet = mat[a] * calcdet(tmp, subdepf);
            if(a & 1)
                sum -= subdet;
            else
                sum += subdet;
        }
    }
    return sum;
}
#else
/* New non-recursive determinant */
static int calcdet(const vector<int> &mat, unsigned depf)
{
    /* mat.size() should be depf * depf */
    
    vector<unsigned> ps(depf); /* position stack   */
    vector<unsigned> pc(depf); /* position counter */
    
    int result = 0, thisresult;

    unsigned level=0, maxlevel=depf-1;
    ps[0] = pc[0] = 0;
    
    for(;;)
    {
        for(unsigned a=0; a<level; ++a)
            if(ps[a] == ps[level])
                goto Rotate;

        if(level < maxlevel)
        {
            ++level;
            ps[level] = 0;
            pc[level] = 0;
            continue;
        }

       {
        bool nega = false;
        for(unsigned a=0; a<depf; ++a)
            if(pc[a]&1)
                nega = !nega;
        
        #if 0
        cerr << ' ' << (nega ? '-' : '+') << ' ';
        for(unsigned a=0; a<depf; ++a)
            cerr << char('A' + a*depf + ps[a]);
        #endif
    
        thisresult = (int) (nega ? -1 : 1);
       }
        for(unsigned a=0; a<depf; ++a)
            thisresult *= mat[a*depf + ps[a]];
        result += thisresult;
        
        while(++pc[level] >= depf-level)
        {
            if(!level)
            {
                /* All done! */
                #if 0
                cerr << endl << endl;
                #endif
                return result;
            }
            --level;
        }
        #if 0
        if(level <= (depf<4 ? 0 : depf-4))cerr << endl;
        #endif
    Rotate:
        ps[level] = (ps[level] + 1) % depf;
    }
}
#endif


static int dotcount=0;
static void dotter()
{
    static unsigned dotcounter=0;
    if(++dotcounter > 5000)
    {
        if(!dotcount)
        {
            fprintf(stderr, "\tPlease wait..");
        }
        ++dotcount;
        dotcounter=0;
        putc('.', stderr);
        fflush(stderr);
    }
}

static void GenHashopFetch(ostream &out, const hashop &op, const string &arrname)
{
    if(op.istable)
    {
        if(op.dec) out << '(';
        out << arrname << "[n";
        if(op.first) { if(op.first>0)out << '+'; out << op.first; }
        out << ']';
        if(op.dec) out << '+' << op.dec << ')';
    }
    else
    {
        const vector<hashmath> &guess = op.components;
        
        out << '(';
        
        bool first = true;
        
        /* Dump negatives (integer<0) not first */
        for(int nega=0; nega<2; ++nega)
            for(unsigned b=0; b<guess.size(); ++b)
            {
                if((guess[b].integer < 0) == !nega)continue;
                
                if(guess[b].exponent == 0)
                {
                    /* Don't display constant "0" unless it's all we have */
                    if(guess[b].integer || guess.size()==1)
                    {
                        if(first) out << guess[b].integer;
                        else out << (guess[b].integer<0 ? " - " : " + ") << abs(guess[b].integer);
                        first = false;
                    }
                }
                else if(guess[b].integer)
                {
                    if(guess[b].integer == -1)
                    {
                        /* -n */
                        if(first)out << "-"; else out << " - ";
                    }
                    else if(guess[b].integer == 1)
                    {
                        if(!first) out << " + ";
                    }
                    else /* abs(integer) > 1 */
                    {
                        if(first) out << ' ' << guess[b].integer;
                        else out << (guess[b].integer<0 ? " - " : " + ") << abs(guess[b].integer);
                        out << '*';
                    }
                    for(unsigned c=0; c<guess[b].exponent; ++c)
                    {
                        if(c)out << '*';
                        out << 'n';
                    }
                    first = false;
                }
            }
        out << ')';
    }
}

static void BinaryGeneration(
    ostream &out,
    const unsigned indlevel,
    const string &arrname,
    const int firstval,
    const hashop *first,
    const hashop *last,
    const char intr=0)
{
    if(first >= last)
    {
        if(intr) out << intr << ' ';
        GenHashopFetch(out, *first, arrname);
        
        out << "/*" << first->last-firstval << "*/";
        
        return;
    }
    
    out << '\n';
    for(unsigned a=0; a<indlevel; ++a)out << ' ';
    
    int lastval=last->last;
    int middleval=(firstval+lastval)/2;
    const hashop *middle;
    for(middle = first
       ;middle < last &&
        abs((int)middle->last - middleval) > abs((int)(middle+1)->last - middleval)
       ;++middle);
    
    if(intr) out << intr << ' ';
    out << "(n<" << middle->last << ')';
    
    BinaryGeneration(out, indlevel+2, arrname, firstval, first, middle, '?');
    
    if(middle+1 >= last)
    {
        out << '\n';
        unsigned b = 0;
        if(intr)b += 2;
        if(first == middle)
        {
            b += 4+1;
            for(unsigned c=10; middle->last >= c; c*=10)++b;
        }
        for(unsigned a=0; a<indlevel+b; ++a)out << ' ';
    }
    
    BinaryGeneration(out, indlevel+2, arrname, middle->last, middle+1, last, ':');
}

static void GenHashOps(
    vector<hashop> &ops,
    const vector<rep> &reps,
    unsigned limit)
{
    /* Previous value */
    unsigned prev = 0;
    
    /* How many bytes have been skipped
     * from the table so far by the ifs
     */
    int indminus = 0;
    
    /* Step 1: Generate the stripes */
    
    for(unsigned a=0; a<reps.size(); ++a)
    {
        hashop op;
        
        if(reps[a].keystart > prev)
        {
            op.istable = true;
            op.last    = reps[a].keystart;
            op.first   = -indminus;
            
            ops.push_back(op);
            
            /* "prev" is now up to date */
            prev = reps[a].keystart;
        }
        
        op.istable = false;
        op.last    = reps[a].keystart + reps[a].len;
        
        op.components = reps[a].components;
        
        ops.push_back(op);
        
        prev = reps[a].keystart + reps[a].len;
        indminus += reps[a].len;
    }
    
    if(prev < limit)
    {
        hashop op;
        op.istable = true;
        op.last    = limit;
        op.first   = -indminus;
        ops.push_back(op);
    }
}

static void genhash(
    ofstream &cfile, const string &name, vector<unsigned> &keys,
    const string &classname="")
{
    //cerr << "genhash(), classname='" << classname << "'\n";
    
    vector<hashop> ops;
    vector<rep> reps;

    unsigned size = keys.size();
        
    for(unsigned minreplen = 2; minreplen <= 1000U; minreplen += min(minreplen/2, 8U))
    {
        reps.clear();
        
        for(unsigned a=0; a<size-1; ++a)
        {
            rep tmp;
            
            tmp.keystart = a;
            tmp.len = 0;
            
            unsigned fails=0;
            for(unsigned len=minreplen; len <= size-a && fails<5; ++len)
            {
                ++fails;
                for(unsigned elemcount=1; elemcount<=len && elemcount <= 2; ++elemcount)
                {
                    dotter();
                    
                    vector<hashmath> guess(elemcount);
                    
                    vector<int> mat(elemcount*elemcount);
                    for(unsigned ma=0; ma<elemcount; ++ma)
                        for(unsigned mb=0; mb<elemcount; ++mb)
                        {
                            unsigned val = 1;
                            for(unsigned c=0; c<mb; ++c)val *= (a+ma);
                            mat[ma*elemcount + mb] = val;
                        }
                    int det = calcdet(mat, elemcount);
                    for(unsigned ma=0; ma<elemcount; ++ma)
                    {
                        vector<int> mat2 = mat;
                        for(unsigned c=0; c<elemcount; ++c)
                            mat2[c*elemcount + ma] = keys[a+c];
                        guess[elemcount-1-ma].integer = calcdet(mat2, elemcount) / det;
                        guess[elemcount-1-ma].exponent= ma;
                    }
                    
                    /* Test the guess (first elemcount elements are known to match) */
                    unsigned ma;
                    for(ma=elemcount; ma <= size-a; ++ma)
                    {
                        int val=a+ma;
                        unsigned result=0;
                        for(unsigned mb=0; mb<guess.size(); ++mb)
                        {
                            int thisval = 1;
                            for(unsigned c=0; c<guess[mb].exponent; ++c)thisval *= val;
                            thisval *= guess[mb].integer;
                            result += thisval;
                        }
                        if(ma < len && result != keys[a+ma])goto GuessWrong;
                        if(result != keys[a+ma])break;
                    }
                    if(ma > len)len = ma;
                    /*
                    */
                    
                    tmp.components = guess;
                    tmp.len = len;
                    fails = 0;
                    break;
                    
                GuessWrong: ;
                }
            }
            
            if(tmp.len >= minreplen)
            {
                /*
                fprintf(stderr,
                    "pos %u, len=%u (min %u), elemcount=%u, guessed!\n   ",
                    a, tmp.len, minreplen, tmp.components.size());
                for(unsigned mb=0; mb<tmp.components.size(); ++mb)
                {
                    fprintf(stderr, "  %+d", tmp.components[mb].integer);
                    for(unsigned c=0; c<tmp.components[mb].exponent; ++c)fprintf(stderr, "x");
                }
                fprintf(stderr, "\n\n");
                */

                reps.push_back(tmp);
                a += tmp.len-1;
            }
        }

        ops.clear();
        GenHashOps(ops, reps, size);
        
        if(ops.size() <= max(8u, keys.size() / 40)) break;
    }
    
    for(unsigned a = reps.size(); a-- > 0; )
    {
        keys.erase(keys.begin() + reps[a].keystart,
                   keys.begin() + reps[a].keystart + reps[a].len);
    }

#if 1
    for(unsigned a=0, b=0; a<ops.size(); b=ops[a++].last)
    {
        if(ops[a].istable)
        {
            unsigned len = ops[a].last-b;
            unsigned eka = b+ops[a].first;
            
            if(len == 1)
            {
                ops[a].istable = false;
                ops[a].components.resize(1);
                ops[a].components[0] = hashmath(keys[eka]);
                
                keys.erase(keys.begin() + eka, keys.begin() + eka + 1);
                for(unsigned c=a+1; c<ops.size(); ++c)
                    if(ops[c].istable)
                        --ops[c].first;
            }
        }
    }
#endif

    bool needtable = false;
    for(unsigned a=0; a<ops.size(); ++a)
        if(ops[a].istable)
            needtable = true;
    bool inlinelookup = !needtable && (ops.size() <= 6);

    const string helpertablename = name + "_tab";
    
    if(classname.size())
    {
        cfile << "const unsigned " << classname << "::size() const { return " << size << "; }\n";
        
        /* At this point, cout has the at() function header */
        
        if(inlinelookup)
        {
            /* at() function body inline */
            cout << " { return ";
            BinaryGeneration(cout, 12, helpertablename, 0, &ops.front(), &ops.back());
            cout << "; }\n";
        }
        else
        {
            cfile << "const unsigned short "
                  << classname
                  << "::at(unsigned n)\n"
                     "{\n";
            
            /* at() function body follows */
            cout << ";\n";
        }
    }
    else
    {
        cfile << "class " << name << "\n"
                "{\n"
                "public:\n"
                "    static inline const unsigned size() { return " << size << "; }\n"
                "    static const unsigned short at(unsigned n)";
        if(inlinelookup)
        {
            cfile << " { return ";
            BinaryGeneration(cfile, 12, helpertablename, 0, &ops.front(), &ops.back());
            cfile << "; }\n"
                     "};\n";
        }
        else
        {
            cfile << ";\n"
                     "};\n"
                     "const unsigned short "
                  << name
                  << "::at(unsigned n)\n"
                     "{\n";
        }
    }
    
    /* At this point, cfile contains the at() function body open. */
    if(needtable)
    {
        unsigned biggestdiff = 0, biggest = 0;
        
#if 1 /* If decrement the values happily */
        bool hope=true, need=false;
        for(unsigned a=0, b=0; a<ops.size(); b=ops[a++].last)
        {
            if(ops[a].istable)
            {
                ops[a].dec = 0;
                
                unsigned len = ops[a].last-b;
                unsigned eka = b+ops[a].first;
                
                unsigned big = 0;
                for(unsigned c=0; c<len; ++c)if(keys[c+eka] > big)big = keys[c+eka];
                unsigned small = big;
                for(unsigned c=0; c<len; ++c)if(keys[c+eka] < small)small = keys[c+eka];
                
                cfile << "    /* Minmax for " << eka << ".." << eka+len-1
                      << ": " << small << ", " << big << " */\n";

                if(len == 1)continue;
                
                if(big > biggest) biggest = big;            
                if(big-small > biggestdiff)
                    biggestdiff = big-small;
                if(big-small > 256)
                    hope = false;
                else if(big >= 256 && big-small <= 256)
                {
                    ops[a].dec = small;
                    need = true;
                }
            }
        }
        
        if(!hope || !need)
        {
            for(unsigned a=0; a<ops.size(); ++a) ops[a].dec = 0;
            /*
            if(biggest >= 1000 && biggestdiff < 1000)
                for(unsigned a=0; a<ops.size(); ++a) ops[a].dec = 1000-biggestdiff;
            else if(biggest >= 10000 && biggestdiff < 10000)
                for(unsigned a=0; a<ops.size(); ++a) ops[a].dec = 10000-biggestdiff;
            */
        }
        
        for(unsigned a=0, b=0; a<ops.size(); b=ops[a++].last)
        {
            if(ops[a].istable && ops[a].dec)
            {
                unsigned len = ops[a].last-b;
                unsigned eka = b+ops[a].first;
                cfile << "    /* Values " << eka << ".." << eka+len-1
                      << " decremented here by " << ops[a].dec
                      << " */\n";
                for(unsigned c=0; c<len; ++c)
                    keys[c+eka] -= ops[a].dec;
            }
        }
#endif

        biggest = 0;
        for(unsigned c=0; c<keys.size(); ++c)if(keys[c] > biggest)biggest = keys[c];
        
        if(keys.size())
        {
            unsigned wid = 6;
            cfile << "    static const unsigned ";
            if(biggest < 256)
            {
                cfile << "char ";
                wid = (biggest >= 100) ? 4 : 3;
            }
            else if(biggest < 65536)
            {
                cfile << "short ";
                wid = (biggest >= 10000) ? 6 : 5;
            }
            cfile << helpertablename << '[' << keys.size() << "] =\n    {\n    ";

            unsigned k=0;
            for(unsigned a=0; a<keys.size(); ++a)
            {
                if(!k)cfile << "\t";
                
                char Buf[16];
                sprintf(Buf, "%u,", keys[a]);
                unsigned l = strlen(Buf);
                
                cfile << Buf;
                while(l < wid) { cfile << ' '; ++l; }
                
                if(++k >= 65/wid) { cfile << "\n    "; k=0; }
            }
            if(k)cfile << "\n";
            cfile << "    };\n";
        }
    } /* End of table output code */
    
    if(!inlinelookup)
    {
        /* At this point, cfile is ready to receive the at() function end. */
        cfile << "    return ";
        
#if 0
        for(unsigned a=0; a<ops.size(); ++a)
        {
            cfile << "\n      " << (a ? ':' : ' ');
            if(a+1 < ops.size())
                cfile << " (n < " << ops[a].last << ") ? ";
            GenHashopFetch(cfile, ops[a], helpertablename);
        }
#else
        BinaryGeneration(cfile, 8, helpertablename, 0, &ops.front(), &ops.back());
#endif

        /* End at() function. */
        cfile << ";\n"
                 "}\n";
    }
}

static const string hashname(unsigned n)
{
    return arrname() + "_" + ((n == srcs.size()) ? dest : srcs[n]);
}

static void listaa(ofstream &cfile,
                   const lista &values,
                   unsigned key, unsigned value)
{
    vector<unsigned> keys(values.size());
    
    string classname;
    classname += (key  ==srcs.size()) ? dest : srcs[key];
    classname += "_to_";
    classname += (value==srcs.size()) ? dest : srcs[value];
    
    for(unsigned a=0; a<values.size(); ++a)
        keys[a] = a;
    
    sort(keys.begin(), keys.end(), verta(values, key));
    
    cout << "\n"
            "class " << classname << " : public charmap\n"
            "{\n"
            "    static const unsigned short at(unsigned n)";
    
    genhash(cfile, classname, keys, classname);
    
    cout << "public:\n"
            "    virtual const unsigned size() const;\n"
            "    virtual const unsigned short getkey(unsigned n) const;\n"
            "    virtual const unsigned short getvalue(unsigned n) const;\n"
            "};\n";

    cfile << "const unsigned short " << classname << "::getkey(unsigned n) const\n"
             "{\n"
             "    return " << hashname(key) << "::at(at(n));\n"
             "}\n"
             "const unsigned short " << classname << "::getvalue(unsigned n) const\n"
             "{\n"
             "    return " << hashname(value) << "::at(at(n));\n"
             "}\n";
}

static void makehash(ofstream &cfile, const lista &values, unsigned ind)
{
    vector<unsigned> key(values.size());
    for(unsigned a=0; a<key.size(); ++a)
        key[a] = values[a].values[ind];

    genhash(cfile, hashname(ind), key, "");
}

/*
    Parameters:
        1 = encoding
        2 = filename (without .h)
        3 = sourceset (or more)
        4 = destset

*/

int main(int argc, const char *const *argv)
{
    dest = argv[argc-1];
    
    for(int a=2; a<argc-1; ++a)srcs.push_back(argv[a]);
    
    valuecount = 1 + srcs.size();
    
    string outcfile = argv[1];
    outcfile += ".cc";
    hfilename = argv[1];
    hfilename += ".h";
    
    ofstream cfile(outcfile.c_str());    

    cout << "\n// See " << outcfile << " for details\n";
    cfile << "#include \"../charmap.hh\"\n"
             "#include \"" << hfilename << "\"\n";
    
    lista values;
    
    for(;;)
    {
        string s;
        getline(cin, s);
        if(!cin.good())break;
        
        if(!s.size())continue;
        
        dotter();

        if(s[0]=='#')
        {
            cout << "//" << s.substr(1) << endl;
            continue;
        }
        if(s[0]=='0')
        {
            unsigned a=0;
            unsigned n=0;
            
            elem tmp;
            
            for(;;)
            {
                if(s[a] == '0' && s[a+1] == 'x')
                {
                    a += 2;
                    unsigned val = 0;
                    while(isxdigit(s[a]))
                    {
                        if(s[a] >= '0' && s[a] <= '9') val=val*16 + (s[a]-'0');
                        else if(s[a] >= 'A' && s[a] <= 'F') val=val*16 + (s[a]+10-'A');
                        else if(s[a] >= 'a' && s[a] <= 'f') val=val*16 + (s[a]+10-'a');
                        ++a;
                    }
                    tmp.values.push_back(val);
                    ++n;
                }
                else if(s[a] == 'x' || isxdigit(s[a]))
                {
                    cout << "Uh oh : '" << s << "'\n";
                }
                
                while(isspace(s[a]))++a;
                
                if(s[a]=='#')
                {
                    while(n < valuecount)
                    {
                        tmp.values.push_back(0xFFFD);
                        ++n;
                    }
                    break;
                }
            }
            
            tmp.desc = s.substr(a+1);
            values.push_back(tmp);
        }
    }
    
    cout << "/*{ " << valuecount << " values:\n";
    
    for(unsigned a=0; a<values.size(); ++a)
    {
        unsigned b = 0;
        
        for(unsigned c=0; c<valuecount; ++c)
        {
            if(c)
            {
                while(b < 7) {cout << ' '; ++b; }
            }
            char Buf[32];
            
            sprintf(Buf, "0x%X,", values[a].values[c]);

            cout << Buf;
            b = strlen(Buf) - 4 ;
        }
        while(b < 7) { cout << ' '; ++b; }
        
        cout << "    //" << values[a].desc << endl;
    }
    cout << "}*/\n\n";
    
    makehash(cfile, values, srcs.size());
    for(unsigned a=0; a<srcs.size(); ++a)
    {
        if(a+MAXFIELDS <= srcs.size())continue;
        
        makehash(cfile, values, a);
    
        listaa(cfile, values, a, srcs.size()); // by src        
        listaa(cfile, values, srcs.size(), a); // by dest
    }
    if(dotcount)
        fprintf(stderr, "\n");
    return 0;
}
